searchData={"items":[{"type":"module","title":"Wongi.Engine","doc":"This is a pure-Elixir forward chaining inference engine based on the classic\nRete algorithm.\n\nIt's derived from an earlier [Ruby library](ruby.md) and has a similar\ninterface.","ref":"Wongi.Engine.html"},{"type":"module","title":"Usage - Wongi.Engine","doc":"The following examples will assume a prelude of\n\n```elixir\nimport Wongi.Engine\nimport Wongi.Engine.DSL\n```\n\nwhich comprises the entire public interface of the library.\n\nFirst, an engine instance needs to be created:\n\n```elixir\nengine = new()\n```","ref":"Wongi.Engine.html#module-usage"},{"type":"module","title":"Knowledge management - Wongi.Engine","doc":"All knowledge in the system is represented as a set of triples in the form of\n`{subject, predicate, object}`. Any complex properties or relationships\nbetween entities can be broken down to this form, in which case the\nrelationship members will take on the role of subjects and objects, and the\ntype of the relationship will be the predicate.\n\nAny Elixir term except the atom `:_` can be used as any element of a triple,\nalthough predicates naturally tend to be atoms.\n\nReferences can be used to naturally represent anonymous graph nodes.\n\nFacts are added into the system with `assert/2` or `assert/4`:\n\n```elixir\nengine = engine |> assert(:earth, :satellite, :moon)\n# or\nengine = engine |> assert({:earth, :satellite, :moon})\n# or\nengine = engine |> assert([:earth, :satellite, :moon])\n# or\nengine = engine |> assert(WME.new(:earth, :satellite, :moon))\n```\n\nWME (working memory element) is the standard Rete term for \"fact\". You would\nrarely need to construct `Wongi.Engine.WME` instances by hand, but you might\nretrieve them from the engine and use in further function calls.\n\nSimilarly, `retract/2` or `retract/4` remove facts from the system.\n\n#","ref":"Wongi.Engine.html#module-knowledge-management"},{"type":"module","title":"Searching - Wongi.Engine","doc":"`select/2` and `select/4` can be used to return a set of facts matching a\ntemplate. A template is a triple where some of the elements can be the special\nplaceholder value `:_`.\n\nAn enumerable of all facts matching the template is returned:\n\n```elixir\n[fact] =\n  engine\n  |> select(:earth, :satellite, :_)\n  |> Enum.to_list()\n\nIO.inspect(fact.object)\n# => :moon\n```","ref":"Wongi.Engine.html#module-searching"},{"type":"module","title":"Rules - Wongi.Engine","doc":"Rules allow expressing more complex conditions than a single template.\n\nA rule is constructed like this:\n\n```elixir\nrule = rule(\"optional name\", forall: [\n  matcher1,\n  matcher2,\n  ...\n])\n\nIO.inspect(rule.ref)\n# => #Reference \n```\n\nThe `ref` field is going to be used later to retrieve the results of rule\nexecution.\n\nThe rule can then be installed into the engine:\n\n```elixir\nengine = engine |> compile(rule)\n```\n\nAlternatively, this form can be used if you don't want an intermediate\nvariable for the rule, although it is less pipeable:\n\n```elixir\n{engine, ref} =\n  engine\n  |> compile_and_get_ref(rule(forall: [...]))\n```\n\nThe `forall` section of a rule consists of a list of matchers (more fully\ndocumented in `Wongi.Engine.DSL`) that express some sort of condition. The\nsimplest matcher is `Wongi.Engine.DSL.has/3` which passes if a fact matching\nits template exists.\n\nA crucial part of matching is the variable bindings. A variable is specified\nusing `Wongi.Engine.DSL.var/1`. The first time a variable is encountered, it\nis bound to the matched value. Subsequent matches will only succeed if the\nvalue is the same as the initially bound one.\n\n`:_` can be used as a placeholder variable that matches anything and is not\nbound to any value.\n\n```elixir\nrule = rule(forall: [\n  has(:_, :satellite, var(:satellite)),\n  has(var(:satellite), :mass, var(:mass))\n])\n\nengine =\n  new()\n  |> compile(rule)\n  |> assert(:earth, :satellite, :moon)\n  |> assert(:moon, :mass, 7.34767309e22)\n```\n\nThe results of rule execution can be retrieved using `tokens/2`, which returns\nan enumerable. A token represents a single possible execution of the matcher\nsequence. Our set of facts satisfies the rule exactly once, so we expect\nexactly one token. The bound variables can then be inspected on it:\n\n```elixir\n[token] = engine |> tokens(rule.ref) |> Enum.to_list()\n\nIO.inspect(token[:satellite])\n# => :moon\nIO.inspect(token[:mass])\n# => 7.34767309e22\n```","ref":"Wongi.Engine.html#module-rules"},{"type":"module","title":"Generation - Wongi.Engine","doc":"In addition to passively examining the results, it is also possible for a rule\nto perform some actions when it is fully satisfied. Generating additional\nfacts is one such action.\n\nFor example, we can add a rule that generates a fact about the gravitational\npull on the satellite:\n\n```elixir\nrule =\n  rule(\n    forall: [\n      has(var(:planet), :satellite, var(:satellite)),\n      has(var(:planet), :mass, var(:planet_mass)),\n      has(var(:satellite), :mass, var(:sat_mass)),\n      has(var(:satellite), :distance, var(:distance)),\n      assign(:pull, &(6.674e-11 * &1[:sat_mass] * &1[:planet_mass] / :math.pow(&1[:distance], 2)))\n    ],\n    do: [\n      gen(var(:satellite), :pull, var(:pull))\n    ]\n  )\n\nengine =\n  new()\n  |> compile(rule)\n  |> assert(:earth, :satellite, :moon)\n  |> assert(:earth, :mass, 5.972e24)\n  |> assert(:moon, :mass, 7.34767309e22)\n  |> assert(:moon, :distance, 384_400.0e3)\n\n[wme] = engine |> select(:moon, :pull, :_) |> Enum.to_list()\nIO.inspect(wme.object)\n# => 1.9819334566450407e20\n```\n\nThe generated facts keep track of the rule that generated them and get\nautomatically retracted if the conditions are no longer satisfied.\n\nIf a fact has been generated by a rule and also asserted manually, it also\nneeds to be retracted by both means to be removed from the system.","ref":"Wongi.Engine.html#module-generation"},{"type":"function","title":"Wongi.Engine.assert/2","doc":"Returns an engine with the given fact added to the working memory.","ref":"Wongi.Engine.html#assert/2"},{"type":"function","title":"Wongi.Engine.assert/4","doc":"Returns an engine with the given fact added to the working memory.","ref":"Wongi.Engine.html#assert/4"},{"type":"function","title":"Wongi.Engine.compile/2","doc":"Returns an engine with the given rule installed.\n\nSee `Wongi.Engine.DSL` for details on the rule definition DSL.","ref":"Wongi.Engine.html#compile/2"},{"type":"function","title":"Wongi.Engine.compile_and_get_ref/2","doc":"Returns an engine with the given rule installed and the rule reference.\n\nThe rule reference can be used to retrieve production tokens using `tokens/2`.\n\nSee `Wongi.Engine.DSL` for details on the rule definition DSL.","ref":"Wongi.Engine.html#compile_and_get_ref/2"},{"type":"function","title":"Wongi.Engine.entity/2","doc":"","ref":"Wongi.Engine.html#entity/2"},{"type":"function","title":"Wongi.Engine.new/0","doc":"Creates a new engine instance.","ref":"Wongi.Engine.html#new/0"},{"type":"function","title":"Wongi.Engine.productions/1","doc":"Returns all production node references.","ref":"Wongi.Engine.html#productions/1"},{"type":"function","title":"Wongi.Engine.retract/2","doc":"Returns an engine with the given fact removed from the working memory.","ref":"Wongi.Engine.html#retract/2"},{"type":"function","title":"Wongi.Engine.retract/4","doc":"Returns an engine with the given fact removed from the working memory.","ref":"Wongi.Engine.html#retract/4"},{"type":"function","title":"Wongi.Engine.select/2","doc":"Returns a set of all facts matching the given template.","ref":"Wongi.Engine.html#select/2"},{"type":"function","title":"Wongi.Engine.select/4","doc":"Returns a set of all facts matching the given template.","ref":"Wongi.Engine.html#select/4"},{"type":"function","title":"Wongi.Engine.tokens/2","doc":"Returns a set of all tokens for the given production node reference.","ref":"Wongi.Engine.html#tokens/2"},{"type":"type","title":"Wongi.Engine.fact/0","doc":"","ref":"Wongi.Engine.html#t:fact/0"},{"type":"opaque","title":"Wongi.Engine.rule/0","doc":"","ref":"Wongi.Engine.html#t:rule/0"},{"type":"opaque","title":"Wongi.Engine.t/0","doc":"","ref":"Wongi.Engine.html#t:t/0"},{"type":"type","title":"Wongi.Engine.template/0","doc":"","ref":"Wongi.Engine.html#t:template/0"},{"type":"opaque","title":"Wongi.Engine.wme/0","doc":"","ref":"Wongi.Engine.html#t:wme/0"},{"type":"module","title":"Wongi.Engine.Aggregates","doc":"Aggregate helpers.","ref":"Wongi.Engine.Aggregates.html"},{"type":"function","title":"Wongi.Engine.Aggregates.max/1","doc":"","ref":"Wongi.Engine.Aggregates.html#max/1"},{"type":"function","title":"Wongi.Engine.Aggregates.min/1","doc":"","ref":"Wongi.Engine.Aggregates.html#min/1"},{"type":"function","title":"Wongi.Engine.Aggregates.product/1","doc":"","ref":"Wongi.Engine.Aggregates.html#product/1"},{"type":"function","title":"Wongi.Engine.Aggregates.sum/1","doc":"","ref":"Wongi.Engine.Aggregates.html#sum/1"},{"type":"module","title":"Wongi.Engine.DSL","doc":"Rule definition functions.","ref":"Wongi.Engine.DSL.html"},{"type":"function","title":"Wongi.Engine.DSL.aggregate/3","doc":"A matcher that computes some value across all its incoming tokens, across all\nexecution paths.\n\nThe matcher produces exactly one token per partition group, unless the\npartition group contains no tokens, in which case the matcher does not pass\nfor that group.","ref":"Wongi.Engine.DSL.html#aggregate/3"},{"type":"function","title":"Wongi.Engine.DSL.any/0","doc":"Placeholder variable. Synonym for `:_`.","ref":"Wongi.Engine.DSL.html#any/0"},{"type":"function","title":"Wongi.Engine.DSL.any/1","doc":"A matcher that passes if any of the sub-chains passes.","ref":"Wongi.Engine.DSL.html#any/1"},{"type":"function","title":"Wongi.Engine.DSL.assign/2","doc":"A matcher that always passes but declares a new variable in the token.\n\nThe value can be constant or a function with arities 0, 1, or 2. A unary\nfunction receives the token as its argument. A binary function receives the\ntoken and the entire engine as its arguments.","ref":"Wongi.Engine.DSL.html#assign/2"},{"type":"function","title":"Wongi.Engine.DSL.diff/2","doc":"A filter that passes if the values are not equal.","ref":"Wongi.Engine.DSL.html#diff/2"},{"type":"function","title":"Wongi.Engine.DSL.equal/2","doc":"A filter that passes if the values are equal.","ref":"Wongi.Engine.DSL.html#equal/2"},{"type":"function","title":"Wongi.Engine.DSL.fact/4","doc":"Synonym for `has/3`, `has/4`.","ref":"Wongi.Engine.DSL.html#fact/4"},{"type":"function","title":"Wongi.Engine.DSL.filter/1","doc":"A filter that passes if the value is a function that returns true.\n\nThe function can have arities 0 or 1. A unary function receives the token as\nits argument.","ref":"Wongi.Engine.DSL.html#filter/1"},{"type":"function","title":"Wongi.Engine.DSL.filter/2","doc":"A filter that passes if the value is a unary function that returns true.\n\nThe function will receive the variable as its argument.","ref":"Wongi.Engine.DSL.html#filter/2"},{"type":"function","title":"Wongi.Engine.DSL.gen/3","doc":"An action that produces new facts.\n\nVariables can be used to refer to values in the token.","ref":"Wongi.Engine.DSL.html#gen/3"},{"type":"function","title":"Wongi.Engine.DSL.greater/2","doc":"A filter that passes if the first value is greater than the second.","ref":"Wongi.Engine.DSL.html#greater/2"},{"type":"function","title":"Wongi.Engine.DSL.gte/2","doc":"A filter that passes if the first value is greater than or equal to the second.","ref":"Wongi.Engine.DSL.html#gte/2"},{"type":"function","title":"Wongi.Engine.DSL.has/4","doc":"A matcher that passes if the specified fact is present in the working memory.\n\n`:_` can be used as a wildcard that matches any value.\n\nVariables can be declared using `var/1`. The first time a template that\ncontains a variable is matched, the variable is bound to the value of the\ncorresponding field in the fact. Subsequent matches will only succeed if the\nvalue of the field is equal to the bound value.","ref":"Wongi.Engine.DSL.html#has/4"},{"type":"function","title":"Wongi.Engine.DSL.in_list/2","doc":"A filter that passes if the first value is a member of the second, which is an\nenumerable.\n\nThe name is misleading since the second value can be anything that implements\n`Enumerable`, but it is kept for similarity with the Ruby DSL.","ref":"Wongi.Engine.DSL.html#in_list/2"},{"type":"function","title":"Wongi.Engine.DSL.less/2","doc":"A filter that passes if the first value is less than the second.","ref":"Wongi.Engine.DSL.html#less/2"},{"type":"function","title":"Wongi.Engine.DSL.lte/2","doc":"A filter that passes if the first value is less than or equal to the second.","ref":"Wongi.Engine.DSL.html#lte/2"},{"type":"function","title":"Wongi.Engine.DSL.missing/3","doc":"Synonym for `neg/3`.","ref":"Wongi.Engine.DSL.html#missing/3"},{"type":"function","title":"Wongi.Engine.DSL.ncc/1","doc":"A matcher that passes if the entire sub-chain does not pass.","ref":"Wongi.Engine.DSL.html#ncc/1"},{"type":"function","title":"Wongi.Engine.DSL.neg/3","doc":"A matcher that passes if the specified fact is not present in the working\nmemory.\n\nVariables declared inside the neg matcher do not become bound, since the\nexecution will only continue if there were no matching facts; however, if the\nsame variable is used in multiple positions within the same neg template,\nunification will be done on it in the scope of the matcher.","ref":"Wongi.Engine.DSL.html#neg/3"},{"type":"function","title":"Wongi.Engine.DSL.none/1","doc":"Synonym for `ncc/1`.","ref":"Wongi.Engine.DSL.html#none/1"},{"type":"function","title":"Wongi.Engine.DSL.not_in_list/2","doc":"A filter that passes if the first value is not a member of the second, which\nis an enumerable.\n\nThe name is misleading since the second value can be anything that implements\n`Enumerable`, but it is kept for similarity with the Ruby DSL.","ref":"Wongi.Engine.DSL.html#not_in_list/2"},{"type":"function","title":"Wongi.Engine.DSL.rule/2","doc":"","ref":"Wongi.Engine.DSL.html#rule/2"},{"type":"function","title":"Wongi.Engine.DSL.var/1","doc":"Variable declaration.","ref":"Wongi.Engine.DSL.html#var/1"},{"type":"type","title":"Wongi.Engine.DSL.action/0","doc":"","ref":"Wongi.Engine.DSL.html#t:action/0"},{"type":"type","title":"Wongi.Engine.DSL.matcher/0","doc":"","ref":"Wongi.Engine.DSL.html#t:matcher/0"},{"type":"opaque","title":"Wongi.Engine.DSL.rule/0","doc":"","ref":"Wongi.Engine.DSL.html#t:rule/0"},{"type":"type","title":"Wongi.Engine.DSL.rule_option/0","doc":"","ref":"Wongi.Engine.DSL.html#t:rule_option/0"},{"type":"protocol","title":"Wongi.Engine.DSL.Clause","doc":"Rule matcher.","ref":"Wongi.Engine.DSL.Clause.html"},{"type":"function","title":"Wongi.Engine.DSL.Clause.compile/2","doc":"","ref":"Wongi.Engine.DSL.Clause.html#compile/2"},{"type":"type","title":"Wongi.Engine.DSL.Clause.t/0","doc":"All the types that implement this protocol.","ref":"Wongi.Engine.DSL.Clause.html#t:t/0"},{"type":"module","title":"Wongi.Engine.DSL.Has","doc":"Positive assertion.","ref":"Wongi.Engine.DSL.Has.html"},{"type":"type","title":"Wongi.Engine.DSL.Has.option/0","doc":"","ref":"Wongi.Engine.DSL.Has.html#t:option/0"},{"type":"module","title":"Wongi.Engine.DSL.Var","doc":"Variable declaration.","ref":"Wongi.Engine.DSL.Var.html"},{"type":"type","title":"Wongi.Engine.DSL.Var.t/0","doc":"","ref":"Wongi.Engine.DSL.Var.html#t:t/0"},{"type":"module","title":"Wongi.Engine.Entity","doc":"An iterator-like object that represents a set of facts about a single subject.","ref":"Wongi.Engine.Entity.html"},{"type":"function","title":"Wongi.Engine.Entity.fetch/2","doc":"","ref":"Wongi.Engine.Entity.html#fetch/2"},{"type":"function","title":"Wongi.Engine.Entity.new/2","doc":"","ref":"Wongi.Engine.Entity.html#new/2"},{"type":"function","title":"Wongi.Engine.Entity.on/2","doc":"","ref":"Wongi.Engine.Entity.html#on/2"},{"type":"function","title":"Wongi.Engine.Entity.over/2","doc":"","ref":"Wongi.Engine.Entity.html#over/2"},{"type":"protocol","title":"Wongi.Engine.Filter","doc":"Filters tokens based on a predicate.","ref":"Wongi.Engine.Filter.html"},{"type":"function","title":"Wongi.Engine.Filter.pass?/2","doc":"","ref":"Wongi.Engine.Filter.html#pass?/2"},{"type":"type","title":"Wongi.Engine.Filter.t/0","doc":"All the types that implement this protocol.","ref":"Wongi.Engine.Filter.html#t:t/0"},{"type":"module","title":"Wongi.Engine.Token","doc":"An intermediate data structure representing a partial match.","ref":"Wongi.Engine.Token.html"},{"type":"function","title":"Wongi.Engine.Token.fetch/2","doc":"Returns the value of a bound variable.","ref":"Wongi.Engine.Token.html#fetch/2"},{"type":"type","title":"Wongi.Engine.Token.t/0","doc":"","ref":"Wongi.Engine.Token.html#t:t/0"},{"type":"module","title":"Wongi.Engine.WME","doc":"A single fact.","ref":"Wongi.Engine.WME.html"},{"type":"function","title":"Wongi.Engine.WME.fetch/2","doc":"","ref":"Wongi.Engine.WME.html#fetch/2"},{"type":"type","title":"Wongi.Engine.WME.t/0","doc":"","ref":"Wongi.Engine.WME.html#t:t/0"},{"type":"extras","title":"Differences from the Ruby library","doc":"# Differences from the Ruby library\n\n- variables are constructed explicitly rather than using the capitalized symbol\n  convention\n- neg matchers can declare variables, although the variables are only used for\n  unification if they're used more than once within the same template, and do\n  not escape the scope of the matcher\n- `assert` is called `filter`\n- `assuming` clauses are not implemented yet\n- queries are not implemented but may be in the future\n- `maybe`/`optional` matcher is not implemented since it's unclear at the moment\n  whether it actually serves a purpose\n- overlays are not implemented and are not going to be, since the immutable\n  nature of the language makes almost the entire feature redundant\n- self-invalidation detection is not implemented and likely isn't going to be,\n  since it is computationally quite expensive while still detecting only a small\n  set of possible contradictions, which is likely impossible to solve in a\n  general case anyway since it seems to boil down to the halting problem; so be\n  aware that your rules may create infinite loops when you generate facts that\n  invalidate their own creation","ref":"ruby.html"}],"content_type":"text/markdown"}